@@ -5,6 +5,7 @@ DripTorch I/O helper functions
 # Internal imports
 from .templates import QuicFire
 from .errors import *
+from ._version import __version__
 
 # External imports
 from folium import Polygon
@@ -131,20 +132,24 @@ def read_geojson_polygon(geojson: dict) -> Polygon:
     return geometry
 
 
-def write_geojson(geometries: list[BaseGeometry], utm_epsg: int, properties={}, style={}) -> dict:
+def write_geojson(geometries: list[BaseGeometry], src_epsg: int, dst_epsg: int = 4326, properties={},
+                  style={}, elapsed_time=None) -> dict:
     """Write a list of shapely geometries to GeoJSON
 
     Args:
         geometries (list[BaseGeometry]): List of shapely geometries
+        src_epsg (int): EPSG code of the CRS that the spatial data are currently projected in.
+        dst_epsg (int): EPSG code of the CRS that the spatial data will be projected to.
         properties (dict, optional): Properties for each feature. Defaults to {}.
         style (dict, optional): Rendering style applied to all features. Defaults to {}.
+        elapsed_time (float, optional): Time elapsed during the firing operation. Defaults to None.
 
     Returns:
         dict: GeoJSON
     """
 
     # Get a projector instance for inverse projection
-    projector = Projector(utm_epsg, 4326)
+    projector = Projector(src_epsg, dst_epsg)
 
     # Get the names of all the props
     property_names = properties.keys()
@@ -171,15 +176,23 @@ def write_geojson(geometries: list[BaseGeometry], utm_epsg: int, properties={},
         'features': features
     }
 
+    # Add elapsed time to properties if it was provided
+    if elapsed_time is not None:
+        geojson['elapsedTime'] = elapsed_time
+
+    # Add DripTorch version to the GeoJSON
+    geojson['driptorchVersion'] = __version__
+
     return geojson
 
 
-def write_quicfire(geometry: list, times: list) -> str:
+def write_quicfire(geometry: list, times: list, elapsed_time, resolution: int = 1) -> str:
     """Writes a QUIC-fire ignition file
 
     Args:
         geometry (list): List of geometry to write
         times (list): Arrival times corresponding to the geometry coordinates (seconds)
+        resolution (int): Horizontal resolution of QUIC-fire domain (meters). Defaults to 1.
 
     Raises:
         ExportError: Error if Point and (Multi)LineString geometry types are mixed
@@ -191,34 +204,49 @@ def write_quicfire(geometry: list, times: list) -> str:
     rows = ''
     n_rows = 0
 
-    # TODO: #20 Cleanup QF export method
+    # Process ignition paths for QF format 5
     if all(isinstance(x, (LineString, MultiLineString)) for x in geometry):
+
+        # Loop over each geometry
         for i, geom in enumerate(geometry):
             time = times[i]
+            # Check if we have a linestring and wrap in a list if so
             if isinstance(geom, LineString):
                 geom = [geom]
                 time = [time]
+            # Loop over each line in the geometry
             for j, part in enumerate(geom):
                 coords = np.array(part.coords)
                 t = time[j]
+                # Loop over each coordinate in the line
                 for k, xy in enumerate(coords[:-1]):
-                    rows += f'{xy[0]} {xy[1]} {coords[k+1,0]} {coords[k+1,1]} {t[k]} {t[1]}\n'
+                    rows += f'{xy[0]} {xy[1]} {coords[k+1,0]} {coords[k+1,1]} {t[k]} {t[k+1]}\n'
                     n_rows += 1
-        file = QuicFire.fmt_5.substitute(n_rows=n_rows, rows=rows)
+        file = QuicFire.fmt_5.substitute(
+            n_rows=n_rows, rows=rows, elapsed_time=round(elapsed_time, 2))
 
+    # Process ignition paths for QF format 4
     elif all(isinstance(x, (Point, MultiPoint)) for x in geometry):
+        # Loop over each geometry
         for i, geom in enumerate(geometry):
             time = times[i]
+            # Check if we have a point and wrap in a list if so
             if isinstance(geom, Point):
                 geom = [geom]
                 time = [time]
+            # Loop over each point in the geometry
             for j, part in enumerate(geom):
                 xy = np.array(part.coords[0])
-                rows += f'{xy[0]} {xy[1]} {time[j]}\n'
+                rows += f'{int(xy[0]/resolution)} {int(xy[1]/resolution)} {time[j]}\n'
                 n_rows += 1
-        file = QuicFire.fmt_4.substitute(n_rows=n_rows, rows=rows)
+        file = QuicFire.fmt_4.substitute(
+            n_rows=n_rows, rows=rows, elapsed_time=round(elapsed_time, 2))
 
+    # Handle the case where we have mixed geometry types
     else:
         raise ExportError(ExportError.incompatible_line_types)
 
+    # Remove blank lines from file
+    file = '\n'.join([line for line in file.split('\n') if line.strip()])
+
     return file
