@@ -4,11 +4,13 @@ Spatiotemporal patterns and the infamous temporal propagator
 
 # Core imports
 from __future__ import annotations
+import copy
 from time import time as unix_time
 import warnings
 
 # Internal imports
-from .io import write_geojson, write_quicfire
+from .io import Projector, write_geojson, write_quicfire
+from .unit import BurnUnit
 
 # External imports
 import awkward as ak
@@ -23,7 +25,7 @@ pd.options.mode.chained_assignment = None
 
 # Turn off the Shapely deprecation warning about about future removal
 # of the array interface. This happens when a Pandas takes a list of Shapely
-# geometies and casts to a list of ndarray objects. GeoPandas would fix this
+# geometries and casts to a list of ndarray objects. GeoPandas would fix this
 # but it's not worth the added complexity of GDAL C binary deps. This is fine
 # as long as we don't upgrade the Shapely req to v2.
 warnings.filterwarnings("ignore", category=ShapelyDeprecationWarning)
@@ -32,7 +34,7 @@ warnings.filterwarnings("ignore", category=ShapelyDeprecationWarning)
 class Pattern:
 
     def __init__(self, heat: list[int], igniter: list[int], leg: list[int],
-                 times: list[list[float]], geometry: list[LineString], utm_epsg: int):
+                 times: list[list[float]], geometry: list[LineString], epsg: int):
         """Constructor
 
         Args:
@@ -41,7 +43,7 @@ class Pattern:
             leg (list[int]): Leg for the path
             times (list[list[float]]): Coordinate arrival times for the path
             geometry (list[LineString]): Path geometry
-            utm_epsg (int): UTM EPSG code for the CRS that the paths are projected in.
+            epsg (int): EPSG code for path geometry
         """
 
         self.heat = heat
@@ -49,16 +51,22 @@ class Pattern:
         self.leg = leg
         self.times = times
         self.geometry = geometry
-        self.utm_epsg = utm_epsg
+        self.epsg = epsg
+
+        # Compute the total elapsed time for the ignition crew
+        times_ak = ak.Array(self.times)
+        min_time = ak.min(times_ak)
+        max_time = ak.max(times_ak)
+        self.elapsed_time = max_time - min_time
 
     @classmethod
-    def from_dict(cls, paths_dict: dict, utm_epsg: int) -> Pattern:
-        """Alternative contructor for initializing a Pattern object with a dictionary
+    def from_dict(cls, paths_dict: dict, epsg: int) -> Pattern:
+        """Alternative constructor for initializing a Pattern object with a dictionary
         of path parameters
 
         Args:
             paths_dict (dict): Dictionary of path parameters
-            utm_epsg (int): UTM EPSG code for the CRS that the paths are currently projected in.
+            burn_unit (BurnUnit): Burn unit object associated with the ignition pattern
 
         Returns:
             Pattern: A new instance of Pattern
@@ -70,7 +78,7 @@ class Pattern:
             paths_dict['leg'],
             paths_dict['times'],
             paths_dict['geometry'],
-            utm_epsg=utm_epsg
+            epsg
         )
 
     def to_dict(self) -> dict:
@@ -110,16 +118,19 @@ class Pattern:
             dict: Timestamped GeoJSON representation of the firing pattern
         """
 
-        # The Timedstamed GeoJSON plug in won't take a time for each coordinate in
+        # Copy the times array
+        times = self.times.copy()
+
+        # The Timedstamed GeoJSON plugin in won't take a time for each coordinate in
         # the sub line string of a MLS, apparently it wants a single time to represent
         # the entire sub line (Either they have a bug or I'm missing something).
         for i, geom in enumerate(self.geometry):
             if isinstance(geom, MultiLineString):
                 # Only keep the start time for each sub line
-                self.times[i] = [time[0] for time in self.times[i]]
+                times[i] = [time[0] for time in times[i]]
 
         # Read the jagged times array as an Awkward array for vectorized operations
-        times = ak.Array(self.times)
+        times = ak.Array(times)
 
         # Convert to milliseconds since Epoch (this is what Leaflet wants)
         times = (times * 1000) + (unix_time() * 1000)
@@ -131,35 +142,87 @@ class Pattern:
             'color': '#ff0000', 'radius': 1}}
 
         # Send off to the GeoJSON writer and return
-        return write_geojson(self.geometry, self.utm_epsg, properties=props, style=style)
+        return write_geojson(self.geometry, self.epsg, properties=props, style=style,
+                             elapsed_time=self.elapsed_time)
 
     def translate(self, x_off: float, y_off: float) -> Pattern:
+        """Translate pattern geometry along the x and y axis by the supplied
+        x and y offset amounts.
 
+        Args:
+            x_off (float): Offset along the x axis
+            y_off (float): Offset along the y axis
+
+        Returns:
+            Pattern: New Pattern object with translated geometries
+        """
+
+        # Translate the path geometries
         geoms = self.geometry
         trans_geoms = []
         for geom in geoms:
             trans_geoms.append(affinity.translate(geom, x_off, y_off))
 
-        self.geometry = trans_geoms
+        # Create a clone of the existing Pattern object and replace the geometries
+        # with the translated geometries
+        obj_copy = copy.copy(self)
+        obj_copy.geometry = trans_geoms
 
-        return self
+        # Return the new Pattern object
+        return obj_copy
 
-    def to_quicfire(self, filename: str = None) -> None | str:
+    def to_quicfire(self, burn_unit: BurnUnit, filename: str = None, time_offset=0, resolution: int = 1, dst_epsg: int = None) -> None | str:
         """Write paths dictionary to QUIC-fire ignition file format.
 
         Args:
+            burn_unit (BurnUnit): Burn unit that defines the extent of the QF simulation
             filename (str, optional): If provided, write the ignition file to the
                 filename. Defaults to None.
+            time_offset (float, optional): Time offset to add to the ignition times. Defaults to 0.
+            resolution (int, optional): Horizontal resolution of QUIC-fire domain (meters). Defaults to 1.
+            dst_epsg (int, optional): EPSG code for the destination projection. Defaults to None.
 
         Returns:
             None | str: None if filename provided, string containing the ignition file if not.
         """
 
+        times = self.times.copy()
+        geometry = self.geometry.copy()
+        domain = burn_unit.copy()
+
+        # Apply time offset if not zero
+        if time_offset:
+            times = ak.Array(times)
+            times = times + time_offset
+            times = times.to_list()
+
+        # Reproject burn_unit and geometry to destination epsg if provided
+        if dst_epsg:
+            projector = Projector(domain.utm_epsg, dst_epsg)
+
+            domain.polygon = projector.forward(domain.polygon)
+
+            reproj_geoms = []
+            for geom in geometry:
+                reproj_geoms.append(projector.forward(geom))
+            geometry = reproj_geoms
+
+        # Translate pattern geometry to the origin or the CRS according to the burn unit extent
+        lower_left = domain.get_bounds().min(axis=0)
+        trans_geoms = []
+        for geom in geometry:
+            trans_geoms.append(affinity.translate(
+                geom, -lower_left[0], -lower_left[1]))
+        geometry = trans_geoms
+
+        # Check if filename was provided and write to it if so
         if filename:
             with open(filename, 'w') as f:
-                f.write(write_quicfire(self.geometry, self.times))
+                f.write(write_quicfire(geometry, times,
+                        self.elapsed_time, resolution=resolution))
+        # Otherwise return QF ignition file to client as string
         else:
-            return write_quicfire(self.geometry, self.times)
+            return write_quicfire(geometry, times, self.elapsed_time, resolution=resolution)
 
 
 class TemporalPropagator:
@@ -168,17 +231,21 @@ class TemporalPropagator:
     through their coordinates.
     """
 
-    def __init__(self, spacing: float, sync_end_time: bool = False):
+    def __init__(self, spacing: float = 0, sync_end_time: bool = False, return_trip: bool = False):
         """Class constructor
 
         Args:
-            spacing (float): Stagger spacing between igniters in a heat (meters)
+            spacing (float): Stagger spacing between igniters in a heat (meters). Defaults to 0.
             sync_end_time (bool, optional): If true synchronize igniters within
                 a heat to finish simultaneously. Defaults to False.
+            return_trip (bool, optional): If true, incorporate the time it takes for a heat
+                of igniters to return to the opposite side of the burn unit (use only for flank ignition).
+                Defaults to False.
         """
 
         self.spacing = spacing
         self.sync_end_time = sync_end_time
+        self.return_trip = return_trip
 
     def forward(self, paths, ignition_crew):
 
@@ -264,6 +331,10 @@ class TemporalPropagator:
                 # The current igniter's start time is the previous heat max end time
                 path.start_time = prev_heat_max_end_time
 
+                # If return_trip is turned on, then incorporate the return trip before this heat starts
+                if self.return_trip:
+                    path.start_time += path.geometry.length / velocity
+
             # This check is for any igniter following the first one. Note that
             # multileg ignition paths are caught in the first check above, so
             # this always deals with the first leg. We'll start by giving the
